/*
 * psrv.c - Process Image server
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "xmalloc.h"
#include "log.h"
#include "list.h"
#include "tree.h"
#include "prioq.h"
#include "buffer.h"
#include "sock-unix.h"
#include "io.h"
#include "timeout.h"
#include "pi-proto.h"
#include "psrv.h"

/*#define DEBUG_PI_SERVER*/

#ifdef DEBUG_PI_SERVER
#define debug(fmt, arg...) debug(fmt, ##arg)
#else
#define	debug(fmt, arg...) do {} while (0)
#endif

struct peer {
	int			sfd;
	buffer_t		rb;
	buffer_t		wb;
	struct sockaddr_un	saddr;
	int connected :1;	/* flag indicates if CMD_CONNECT was issued by client */
};

struct psrv_callbacks callbacks;

static int sockfd = -1;
static slist_t peer_list = NULL;
static struct io_context *io = NULL;

static int side;
static size_t inputs_len;
static void *inputs;
static size_t outputs_len;
static void *outputs;

static void server_event(int fd, short events, void *opaque);

/*
void dump_io()
{
	int i;
	unsigned char *data;

	if (inputs) {
		printf("inputs\n");
		data = inputs;
		for (i = 0; i <inputs_len; i++) {
			if (i % 8) {
				printf("0x%02x ", data[i]);
			} else {
				printf("\n");
				printf("%04x:%04x 0x%02x ", i&0xffff, i>>16, data[i]);
			}
		}
		printf("\n");
	}
	if (outputs) {
		printf("outputs\n");
		data = outputs;
		for (i = 0; i <outputs_len; i++) {
			if (i % 8) {
				printf("0x%02x ", data[i]);
			} else {
				printf("\n");
				printf("%04x:%04x 0x%02x ", i&0xffff, i>>16, data[i]);
			}
		}
		printf("\n");
	}
}
*/

static struct peer *peer_open(int fd, struct sockaddr_un *saddr)
{
	struct peer *peer;
	slist_t list;

	debug("opening new peer session fd=%d", fd);

	peer = xmalloc0(sizeof(*peer));
	peer->sfd = fd;
	peer->rb = buffer_new();
	peer->wb = buffer_new();
	peer->saddr = *saddr;
	sock_nonblock(peer->sfd);
	io_add(io, fd, EVENT_READ, server_event, peer);
	list = slist_new();
	list->data = peer;
	peer_list = slist_add(peer_list, list);
	return peer;
}

static void peer_close(struct peer *peer)
{
	slist_t list;

	debug("closing peer session fd=%d", peer->sfd);

	list = slist_find(peer_list, peer, NULL);
	if (list)
		peer_list = slist_unlink(peer_list, list);
	if (peer->connected &&
	    callbacks.client_disconnect_callback != NULL)
		(*callbacks.client_disconnect_callback)();
	buffer_free(peer->rb);
	buffer_free(peer->wb);
	close(peer->sfd);
	io_del(io, peer->sfd);
	slist_free(&list);
	xfree(peer);
}

static void peer_close_all()
{
	slist_t iter, next = NULL;
	struct peer *peer;

	iter = peer_list;
	while (iter != NULL) {
		next = iter->next;
		peer = iter->data;
		buffer_free(peer->wb);
		buffer_free(peer->rb);
		close(peer->sfd);
		io_del(io, peer->sfd);
		slist_free(&iter);
		xfree(peer);
		iter = next;
	}
}

static void status_send(int fd, int cmd, int status, struct peer *peer)
{
	buffer_put_u16(peer->wb, 2);
	buffer_put_u8(peer->wb, cmd);
	buffer_put_u8(peer->wb, status);
	buffer_write_fd(fd, peer->wb);
}

static void status_send_range(int fd, int cmd, int status,
		int len, int offset, struct peer *peer)
{
	buffer_put_u16(peer->wb, 6);
	buffer_put_u8(peer->wb, cmd);
	buffer_put_u8(peer->wb, status);
	buffer_put_u16(peer->wb, len);
	buffer_put_u16(peer->wb, offset);
	buffer_write_fd(fd, peer->wb);
}

static void config_send(int fd, struct peer *peer)
{
	buffer_put_u16(peer->wb, 8);
	buffer_put_u8(peer->wb, CMD_OBTAIN_CONFIG);
	buffer_put_u8(peer->wb, STATUS_OK);
	buffer_put_u16(peer->wb, inputs_len);
	buffer_put_u16(peer->wb, outputs_len);
	if (side == SIDE_MASTER)
		buffer_put_u16(peer->wb, SIDE_MASTER);
	else if (side == SIDE_SLAVE)
		buffer_put_u16(peer->wb, SIDE_SLAVE);
	else
		buffer_put_u16(peer->wb, 0);
	buffer_write_fd(fd, peer->wb);
}

static void packet_receive(int fd, struct buffer *pkt, struct peer *peer)
{
	u_int8_t cmd;
	u_int8_t status;
	u_int16_t plen;
	size_t len;
	size_t offset;

	if (buffer_size(pkt) < 4)
		return;

	plen = buffer_get_u16(pkt);
	cmd = buffer_get_u8(pkt);
	status = buffer_get_u8(pkt);

	switch (cmd) {
	case CMD_CONNECT:
		/* send confirmation */
		debug("sending CMD_CONNECT response");
		if (callbacks.client_connect_callback) {
			if ((*callbacks.client_connect_callback)() != 0)
				status_send(fd, CMD_CONNECT, STATUS_NOT_AVAIL, peer);
			else
				status_send(fd, CMD_CONNECT, STATUS_OK, peer);
		} else
			status_send(fd, CMD_CONNECT, STATUS_OK, peer);
		peer->connected = 1;
		break;
	case CMD_OBTAIN_CONFIG:
		if (peer->connected) {
			debug("sending CMD_OBTAIN_CONFIG response");
			config_send(fd, peer);
		} else {
			/* sorry my friend, but you are not connected */
			status_send(fd, CMD_OBTAIN_CONFIG,
					STATUS_NOT_CONNECTED, peer);
		}
		break;
	case CMD_PREPARE_READ:
		if (!peer->connected) {
			debug("sending CMD_PREPARE_READ NOT_CONNECTED");
			status_send(fd, CMD_PREPARE_READ,
					STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (callbacks.prepare_read_callback) {
			if ((*callbacks.prepare_read_callback)() != 0)
				status_send(fd, CMD_PREPARE_READ,
						STATUS_NOT_AVAIL, peer);
			else
				status_send(fd, CMD_PREPARE_READ,
						STATUS_OK, peer);
		} else
			status_send(fd, CMD_PREPARE_READ, STATUS_OK, peer);
		debug("sending CMD_PREPARE_READ response");
		break;
	case CMD_PREPARE_WRITE:
		if (!peer->connected) {
			debug("sending CMD_PREPARE_WRITE NOT_CONNECTED");
			status_send(fd, CMD_PREPARE_WRITE,
					STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (callbacks.prepare_write_callback) {
			 if ((*callbacks.prepare_write_callback)() != 0)
				 status_send(fd, CMD_PREPARE_WRITE,
						 STATUS_NOT_AVAIL, peer);
			 else
				 status_send(fd, CMD_PREPARE_WRITE,
						 STATUS_OK, peer);
		} else
			status_send(fd, CMD_PREPARE_WRITE, STATUS_OK, peer);
		debug("sending CMD_PREPARE_WRITE response");
		break;
	case CMD_END_READ:
		if (!peer->connected) {
			debug("sending CMD_END_READ NOT_CONNECTED");
			status_send(fd, CMD_END_READ, STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (callbacks.end_read_callback) {
			if ((*callbacks.end_read_callback)() != 0)
				status_send(fd, CMD_END_READ, STATUS_NOT_AVAIL, peer);
			else
				status_send(fd, CMD_END_READ, STATUS_OK, peer);
		} else
			status_send(fd, CMD_END_READ, STATUS_OK, peer);
		debug("sending CMD_END_READ response");
		break;
	case CMD_END_WRITE:
		if (!peer->connected) {
			debug("sending CMD_END_WRITE NOT_CONNECTED");
			status_send(fd, CMD_END_WRITE, STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (callbacks.end_write_callback) {
			if ((*callbacks.end_write_callback)() != 0)
				status_send(fd, CMD_END_WRITE, STATUS_NOT_AVAIL, peer);
			else
				status_send(fd, CMD_END_WRITE, STATUS_OK, peer);
		} else
			status_send(fd, CMD_END_WRITE, STATUS_OK, peer);
		debug("sending CMD_END_WRITE response");
		break;
	case CMD_INPUTS_READ:
		if (!peer->connected) {
			debug("sending CMD_INPUTS_READ NOT_CONNECTED");
			status_send(fd, CMD_INPUTS_READ,
					STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (buffer_size(pkt) >= 4) {
			len = buffer_get_u16(pkt);
			offset = buffer_get_u16(pkt);
			if (offset + len <= inputs_len && len <= PEER_MAX_MSG-6) {
				buffer_put_u16(peer->wb, len+6);
				buffer_put_u8(peer->wb, CMD_INPUTS_READ);
				buffer_put_u8(peer->wb, STATUS_OK);
				buffer_put_u16(peer->wb, len);
				buffer_put_u16(peer->wb, offset);
				buffer_put(peer->wb, inputs + offset, len);
				debug("sending INPUTS_READ=OK %lu at %lu", (unsigned long)len, (unsigned long)offset);
				buffer_write_fd(fd, peer->wb);
			} else if (inputs_len > 0) {
				debug("sending INPUTS_READ=ERANGE %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_INPUTS_READ,
						STATUS_ERANGE, len, offset, peer);
			} else {
				debug("sending INPUTS_READ=NOT_AVAIL %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_INPUTS_READ,
						STATUS_NOT_AVAIL, len, offset, peer);
			}
		} else {
			debug("sending INPUT_READ=MALFORMED");
			status_send(fd, CMD_INPUTS_READ, STATUS_MALFORMED, peer);
		}
		break;
	case CMD_OUTPUTS_READ:
		if (!peer->connected) {
			debug("sending CMD_OUTPUTS_READ NOT_CONNECTED");
			status_send(fd, CMD_OUTPUTS_READ, STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (buffer_size(pkt) >= 4) {
			len = buffer_get_u16(pkt);
			offset = buffer_get_u16(pkt);
			if (offset + len <= outputs_len && len <= PEER_MAX_MSG-6) {
				buffer_put_u16(peer->wb, len+6);
				buffer_put_u8(peer->wb, CMD_OUTPUTS_READ);
				buffer_put_u8(peer->wb, STATUS_OK);
				buffer_put_u16(peer->wb, len);
				buffer_put_u16(peer->wb, offset);
				buffer_put(peer->wb, outputs + offset, len);
				debug("sending OUTPUTS_READ=OK %lu at %lu", (unsigned long)len, (unsigned long)offset);
				buffer_write_fd(fd, peer->wb);
			} else if (outputs_len > 0) {
				debug("sending OUTPUTS_READ=ERANGE %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_OUTPUTS_READ,
						STATUS_ERANGE, len, offset, peer);
			} else {
				debug("sending OUTPUTS_READ=NOT_AVAIL %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_OUTPUTS_READ,
						STATUS_NOT_AVAIL, len, offset, peer);
			}
		} else {
			debug("sending OUTPUTS_READ=MALFORMED");
			status_send(fd, CMD_OUTPUTS_READ, STATUS_MALFORMED, peer);
		}
		break;
	case CMD_OUTPUTS_WRITE:
		if (!peer->connected) {
			debug("sending CMD_OUTPUTS_WRITE NOT_CONNECTED");
			status_send(fd, CMD_OUTPUTS_WRITE, STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (buffer_size(pkt) >= 4) {
			len = buffer_get_u16(pkt);
			offset = buffer_get_u16(pkt);
			if (buffer_size(pkt) == len && len + offset <= outputs_len) {
				/* Not safe ! */
				memcpy(outputs + offset, buffer_ptr(pkt), len);
				debug("sending OUTPUTS_WRITE=OK %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_OUTPUTS_WRITE,
						STATUS_OK, len, offset, peer);
			} else if (outputs_len > 0) {
				debug("sending OUTPUTS_WRITE=ERANGE %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_OUTPUTS_WRITE,
						STATUS_ERANGE, len, offset, peer);
			} else {
				debug("sending OUTPUTS_WRITE=NOT_AVAIL %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_OUTPUTS_WRITE,
						STATUS_NOT_AVAIL, len, offset, peer);
			}
		} else {
			debug("sending OUTPUTS_WRITE=MALFORMED");
			status_send(fd, CMD_OUTPUTS_WRITE, STATUS_MALFORMED, peer);
		}
		break;
	case CMD_INPUTS_WRITE:
		if (!peer->connected) {
			debug("sending CMD_INPUTS_WRITE NOT_CONNECTED");
			status_send(fd, CMD_INPUTS_WRITE, STATUS_NOT_CONNECTED, peer);
			return;
		}
		if (buffer_size(pkt) >= 4) {
			len = buffer_get_u16(pkt);
			offset = buffer_get_u16(pkt);
			if (buffer_size(pkt) == len && len + offset <= inputs_len) {
				memcpy(inputs + offset, buffer_ptr(pkt), len);
				debug("sending INPUTS_WRITE=OK %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_INPUTS_WRITE,
						STATUS_OK, len, offset, peer);
			} else if (inputs_len > 0) {
				debug("sending INPUTS_WRITE=ERANGE %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_INPUTS_WRITE,
						STATUS_ERANGE, len, offset, peer);
			} else {
				debug("sending INPUTS_WRITE=NOT_AVAIL %lu at %lu", (unsigned long)len, (unsigned long)offset);
				status_send_range(fd, CMD_INPUTS_WRITE,
						STATUS_NOT_AVAIL, len, offset, peer);
			}
		} else {
			debug("sending INPUTS_WRITE=MALFORMED");
			status_send(fd, CMD_INPUTS_WRITE, STATUS_MALFORMED, peer);
		}
		break;
	default:
		status_send(fd, cmd, STATUS_BAD_COMMAND, peer);
		break;
	}
}

static void server_event(int fd, short events, void *opaque)
{
	int res;
	u_int16_t len;
	struct peer *peer = (struct peer *) opaque;
	struct buffer *pkt;

	if (events & EVENT_ERROR) {
		peer_close(peer);
		return;
	}

	if (events & EVENT_READ) {
		res = buffer_read_fd(peer->sfd, peer->rb, 4096);
		if (res == 0) {
			peer_close(peer);
			return;
		} else if (res == -1 && (errno != EINTR &&
		     errno != EAGAIN && errno != EWOULDBLOCK)) {
			warn("read()=%d sock=%d: %s", res,
					peer->sfd, strerror(errno));
		} else if (res > 0) {
			pkt = buffer_new();
			while (buffer_size(peer->rb) >= 4) {
				len = buffer_get_u16_at(peer->rb, 0);
				if (len > PEER_MAX_MSG) {
					warn("dropped %d bytes of junk message", len);
					buffer_reset(peer->rb);
					break;
				}
				if (buffer_size(peer->rb) >= len + 2) {
					buffer_copy(pkt, peer->rb, len + 2);
					packet_receive(peer->sfd, pkt, peer);
					buffer_consume(peer->rb, len + 2);
					buffer_reset(pkt);
				} else
					break;
			}
			buffer_free(pkt);
		}
	}

	if (events & EVENT_WRITE) {
		if (buffer_size(peer->wb) > 0) {
			res = buffer_write_fd(peer->sfd, peer->wb);
			if (res == -1 && (errno != EINTR ||
			     errno != EWOULDBLOCK || errno != EAGAIN)) {
				warn("write()=%d sock=%d: %s", res,
						peer->sfd, strerror(errno));
			}
		}
		if (buffer_size(peer->wb) == 0)
			io_change(io, peer->sfd, EVENT_READ);
	}
}

static void accept_event(int fd, short events, void *opaque)
{
	int client_fd;
	socklen_t slen;
	struct sockaddr_un saddr;

	if (events & EVENT_ERROR) {
		/* Oooh... */
		peer_close_all();
		/* exit */
		error_exit(2, "error on listening socket");
	}

	if (events & EVENT_READ) {
		slen = sizeof(saddr);
		client_fd = accept(fd, (struct sockaddr *) &saddr, &slen);
		if (client_fd >= 0) {
			peer_open(client_fd, &saddr);
		} else {
			warn("accept() failed with code=%d: %s",
					client_fd, strerror(errno));
		}
	}
}

int psrv_init(struct io_context *io_ptr, const char *sockfile, void *inptr, size_t isize, void *optr, size_t osize, int master)
{
	int sfd;

	if (!io_ptr || !sockfile)
		error("psrv_init: arg NULL");

	if (io != NULL) {
		warn("psrv_init: already initialized");
		return -1;
	}

	if (master)
		side = SIDE_MASTER;
	else
		side = SIDE_SLAVE;

	sfd = sock_unix_listen(sockfile);
	if (sfd == -1)
		return -1;
	sock_nonblock(sfd);
	io_add(io_ptr, sfd, EVENT_READ, accept_event, NULL);
	sockfd = sfd;
	io = io_ptr;
	if (isize && inptr != NULL) {
		inputs = inptr;
		inputs_len = isize;
	}
	if (osize && optr != NULL) {
		outputs = optr;
		outputs_len = osize;
	}
	return 0;
}

int psrv_register_callbacks(struct psrv_callbacks *new, struct psrv_callbacks *old)
{
	if (!old && !new)
		return -1;

	if (old != NULL)
		*old = callbacks;

	if (new != NULL) {
		if (new->end_read_callback != NULL)
			callbacks.end_read_callback = new->end_read_callback;
		else
			callbacks.end_read_callback = NULL;
		if (new->end_write_callback != NULL)
			callbacks.end_write_callback = new->end_write_callback;
		else
			callbacks.end_write_callback = NULL;
		if (new->prepare_read_callback != NULL)
			callbacks.prepare_read_callback = new->prepare_read_callback;
		else
			callbacks.prepare_read_callback = NULL;
		if (new->prepare_write_callback)
			callbacks.prepare_write_callback = new->prepare_write_callback;
		else
			callbacks.prepare_write_callback = NULL;
	}
	return 0;
}

void psrv_deinit()
{
	if (io == NULL) {
		warn("psrv_deinit: not initialized");
		return;
	}

	peer_close_all();
	io_del(io, sockfd);
	sock_unix_close(sockfd);
	io = NULL;
	inputs_len = 0;
	outputs_len = 0;
	inputs = NULL;
	outputs = NULL;
	sockfd = -1;
	peer_list = NULL;
}
